#!/usr/bin/env python
# encoding: utf-8
"""
txt2hkspc.py

This converts EAERI format text files containing housekeeping data into DMV-formatted *HK.SPC files.

Created by Keoni on 2008-03-24.
Copyright (c) 2008 University of Wisconsin SSEC. All rights reserved.
"""

import sys, os, re, logging, time

# from time_tools import *
import dmv

HISTORY_FMT = """written by $Id: txt2hkspc.py,v 1.1.2.5 2008/05/15 20:30:31 rayg Exp $ from %s"""

LOG = logging.getLogger('txt2hkspc')

# keys into the housekeeping dictionary which are unlabeled in the text file
KEY_YYYYMMDD = 'YYYYMMDD'
KEY_HHMMSS = 'HHMMSS'
KEY_JULIANDAY = 'JulianDay'   # derived from YYYYMMDD
KEY_TIME = 'Time'   # derived from HHMMSS

# nan constant
nan = float('nan')

#
# Summary functions
# reduce the lists of value down to single summary values
#

def summarize_first_value(values):
    return values[0]
    
def summarize_last_value(values):
    return values[-1]
    
def summarize_center_value(values):
    return values[int(len(values)/2)]

def summarize_mean_value(values):
    return sum(values) / float(len(values))
    
def summarize_flag(values):
    "return False for fully off, True for fully on, None for unstable"
    tot = sum([x and 1 or 0 for x in values])
    if tot==len(values): return True
    if tot==0: return False
    return None

def calc_hatch_open(D):
    """assign hatch values given hatchopendetect and hatchcloseddetect similarly to aerihk.cpp:AERIHK::calcHatchOpen
    """
    isOpen = summarize_flag(D['openLimitSwitch'])
    LOG.debug("isOpen: %s" % isOpen)
    isClosed = summarize_flag(D['closeLimitSwitch'])
    LOG.debug("isClosed: %s" % isClosed)
    if isOpen is True and isClosed is True: return -1.
    if isOpen is False and isClosed is True: return 0.
    if isOpen is True and isClosed is False: return 1.
    return -3.
    
def calc_jday(ymdstr):
    """convert a YYYYMMDD string to day-of-year
    """
    yyyy = int(ymdstr[:4])
    mm = int(ymdstr[4:6])
    dd = int(ymdstr[6:8])
    t = time.mktime((yyyy, mm, dd, 0, 0, 0, 0, 0, 0))
    # compute julian day
    jday = time.gmtime(t)[7]
    LOG.debug('julian day is %d' % jday)
    return jday
    
def calc_hours(hhmmss):
    """convert a HHMMSS string to decimal hours, returning a float
    """
    hh = int(hhmmss[0:2])
    mm = int(hhmmss[2:4])
    ss = int(hhmmss[4:6])
    T = float(hh) + float(mm)/60. + float(ss)/3600.
    LOG.debug('decimal hours is %f' % T)
    return T

def read_txt(txt):
    """ reads EAERI .hk format file content given as a text string
    returns a dictionary of value sequences for housekeeping text
    read_txt(eaeri-housekeeping-text-string)
        => { 'name': [ floats ] , ... }
    """
    # read the file and extract list of pairs
    p = re.findall(r'(\w+)[\s=]+([.\d\-]+)', txt) # note assumption that findall is stable
    LOG.debug('found %d pairs in housekeeping text' % len(p))
    # extract date and time
    times = re.findall( r'^(\d{8})_([.\d]+)', txt, re.MULTILINE)    
    LOG.debug("found %d time values in text" % len(times))
    # make a dictionary of name: [list-of-values] in time order
    tab = { }
    for name,val in p:
        tab.setdefault(name,[]).append(float(val))
    tab[KEY_YYYYMMDD] = [int(x[0]) for x in times]
    tab[KEY_JULIANDAY] = [calc_jday(x[0]) for x in times]
    tab[KEY_HHMMSS] = [float(x[1]) for x in times]
    tab[KEY_TIME] = [calc_hours(x[1]) for x in times]
    return tab

# list of shortname, eaeriname, label, longname, blank, units, precision for vectors
# ( 'shortname', 'eaeriname', 'label', 'longname', blank, 'units', -4 )
VECTOR_TABLE = [ ('timeHHMMSS',KEY_HHMMSS,'Time in HHMMSS',None,None,None,0),
            ('dateYYMMDD',KEY_YYYYMMDD,'Date in YYYYMMDD',None,None,None,0),
            ('Time', KEY_TIME, 'Time of day in decimal hours', None,None,None, -4),
            ('JulianDay', KEY_JULIANDAY, 'Julian day', None,None,None, 0),
            
            ('HBBapexTemp', 'hotBlackbodyApexTemperature', 'HBB Apex Temp', None,None, 'K', -3), 
            ('HBBtopTemp', 'hotBlackbodyTopRimTemperature', 'HBB Top Temp', None,None, None, -3),
            ('HBBbottomTemp', 'hotBlackbodyBottomRimTemperature', 'HBB Bottom Temp', None,None, None, -3),
            ('ABBapexTemp', 'ambientBlackbodyApexTemperature', 'HBB Apex Temp', None,None, 'K', -3), 
            ('ABBtopTemp', 'ambientBlackbodyTopRimTemperature', None, None, None,None,-3),
            ('ABBbottomTemp', 'ambientBlackbodyBottomRimTemperature', None, None, None,None,-3),

            ('fixed97KohmResistor', 'refResistance2', None, None, None, None,-3),  # FIXME: 1, 2, or 3?
            ('fixed2500ohmResistor', 'refResistance1', None, None, None, None,-3), # FIXME: 1, 2, or 3?
            ('fixed12KohmResistor', 'refResistance3', None, None, None, None,-3),  # FIXME: 1, 2, or 3?

            ('outsideAirTemp', 'outsideAirTemperature', None, None, None, None,-3),
            ('airNearInterferometerTemp','interferometerAirTemperature','',None, None, None,-3),
            ('mirrorMotorTemp','mirrorMotorTemperature','',None,None, None,-3),            
            ('sceneMirrorTemp','sceneMirrorTemperature','',None,None,None,-3),
            ('interferometerWindowTemp','sceneWindowTemperature','',None,None,None,-3),
            
            ('coolerExpanderTemp','stirlingCoolerExpenderResistance','',None,None,None,-3),
            ('coolerExpanderTemp','stirlingCoolerExpenderTemperature','',None,None,None,-3),  # FIXME misspelling
            ('coolerCompressorTemp','stirlingCoolerCompressorTemperature','',None,None,None,-3),
            ('detectorTemp','detectorTemperatureSensor','',None,None,None,-3),
            ('opticsCompartmentRelativeHumidity','interferometerRelativeHumidity','',None,None,None,-3),

            ('atmosphericRelativeHumidity','atmosphericRelativeHumidity','',None,None,None,-3),
            ('airNearBBsTemp','blackbodyAirTemperature','',None,None,None,-3),
            ('sceneMirrorAngle','pointingMirrorAngle','',None,None,None,-3),
            ('atmosphericPressure','atmosphericPressureSensor','',None,None,None,-3),
            ('sunSensorIntensity','externalSunDetectorSensor','',None,None,None,-3),
            ('rainSensorIntensity','internalRainDetectorSensor','',None,None,None,-3),
            ('SCEtemp','dataAcquisitionModulesTemperature','',None,None,None,-3),

            ('','refTemp1','',None,None,None,-3),
            ('','dataAcquisitionModulesResistance','',None,None,None,-3),
            ('','sceneWindowResistance','',None,None,None,-3),
            ('','blackbodySupportStructureResistance','',None,None,None,-3),
            ('','mirrorMotorResistance','',None,None,None,-3),
            ('','powerSupplyResistance24V','',None,None,None,-3),
            ('coolerCurrent','stirlingCoolerCurrent','',None,None,None,-3),
            ('','inputMirrorMotorDriveElectronicResistance','',None,None,None,-3),
            ('','blackbodyAirResistance','',None,None,None,-3),
            ('','refTemp2','',None,None,None,-3),
            ('','outsideRainSensor','',None,None,None,-3),
            ('','powerSupplyTemperature28V','',None,None,None,-3),
            ('hatchClosedDetect','closeLimitSwitch','',None,None,None,-3),
            ('','externalRainDetectorSensor','',None,None,None,-3),
            ('','thermalEnclosureResistance','',None,None,None,-3),
            ('','powerSupplyResistance28V','',None,None,None,-3),
            ('','outsideAirResistance','',None,None,None,-3),
            ('','protectiveEnclosureResistance','',None,None,None,-3),
            ('sunSensorIntensity','internalSunDetectorSensor','',None,None,None,-3),
            ('','inputMirrorMotorDriveElectronicTemperature','',None,None,None,-3),
            ('','powerSupplyTemperature24V','',None,None,None,-3),
            ('','protectiveEnclosureTemperature','',None,None,None,-3),
            ('','interferometerAirResistance','',None,None,None,-3),
            ('','stirlingCoolerCompressorResistance','',None,None,None,-3),
            ('','sceneMirrorResistance','',None,None,None,-3),
            ('','fanAirflowTemperature','',None,None,None,-3),
            ('','fanAirflowResistance','',None,None,None,-3),
            ('rainSensorIntensity','insideRainSensor','',None,None,None,-3),
            ('','thermalEnclosureTemperature','',None,None,None,-3),
            ('hatchOpenDetect','openLimitSwitch','',None,None,None,-3),
            ]    

def extract_scalars(vectors, table = VECTOR_TABLE):
    """ extract scalars from the high-resolution samples
    return dictionary of name -> float
    refer to augmentListsCalc() in aerihk.cpp
    """
    zult = { }
    for vec_sname, vec_ename, _, _, handler, _, _ in table:
        if not vec_sname: continue
        if not handler: handler = summarize_mean_value
        zult[vec_sname] = handler(vectors[vec_ename])
        
    # now extract values with more obscure handling
    zult['hatchOpen'] = calc_hatch_open(vectors)
    
    LOG.warning( "Not Implemented: " + "assorted housekeeping value computations not available / impossible" )
    
    # zult['sceneViewDuration'] = FIXME    # seconds, actual dwell time
    # zult['maxSampleStdDev'] = FIXME   # maximum standard deviation among selected HK channels, for stability measure
    # zult['maxPitch'] = FIXME  # clinometer max pitch, degrees
    # zult['sceneMirrorMotorStep'] = FIXME  # motor (commanded) step for scene mirror, counts
    # zult['sceneMirPosEncoder'] = FIXME    # encoder reported position value for scene mirror, encoder counts
    # zult['sceneMirPosEncoderDrift'] = FIXME # drift between expected encoder value and actual, encoder counts
    # zult['frontEndFanAirTempDiff'] = FIXME  # difference of heated - unheated front end fan air temps
    # zult['ABBmaxTempDiff'] = FIXME # maximum temp diff between ABB thermistor channels
    # zult['HBBmaxTempDiff'] = FIXME # as well as HBB channels
    # Time and JulianDay are unpacked during file read
    return zult
    
AERI_MODELS = { 'AERI': 1, 
                'MAERI': 2,
                'PAERI': 3,
                'EAERI': 4 
                }
                
def read_inst_id(filename):
    """ return model and serial number tuple given the filename of an inst.id file
    refer aerihk.cpp:AERIexterns::readINST_ID()
    """
    if not os.path.exists(filename):
        LOG.error('INST.ID is missing and should be set to EAERI-###')
        return nan, nan
    txt = file(filename,'rt').read().strip()
    LOG.debug('text read from INST.ID: "%s"' % txt)
    smod,sser = txt.split('-')
    return AERI_MODELS[smod], int(sser)
    
# regular expression which locates the number of coadds (not commented out) in a SET file
RE_COADDS = re.compile(r'^[^#]*smnsca\s*=\s*(\d+)', re.IGNORECASE|re.MULTILINE)

def read_set_coadds(filename):
    """ return coadds from .SET file
    refer aerihk.cpp:AERIexterns::readSetCoadds()
    """
    if not os.path.exists(filename):
        LOG.error('SET file is missing and should be set to EAERI-###')
        return nan
    text = file(filename,'rt').read()
    # note that this will throw a ValueError exception in the case that more than one valid value is found
    (q,) = RE_COADDS.findall(text)
    return int(q) 
    
def read_aeri_pos(filename):
    """ return Latitude N, Longitude E, Altitude meters values
    """
    # FIXME: unimplemented
    LOG.warning("read_aeri_pos Not Implemented")
    return None, None, None
    
def config_scalars(set_filename, inst_id_filename, aeri_pos_filename):
    """ return scalars which need to be read from configuration files
    set_filename is the filename of the .SET file which was used for this ingest cycle, 
    typically a generated file living in the ingest work directory and not a config file
    FUTURE: the Sequencer can provide the number of coadds to this routine instead
    """
    LOG.warning( "Not Implemented: " +"systemReleaseNumber and sceneMirPosCount not implemented" )
    if inst_id_filename:
        model,serial = read_inst_id(inst_id_filename)
    else:
        LOG.error("no INST.ID filename to read model and serial number from")
        model,serial = nan,nan
        
    if not set_filename:
        LOG.error("no SET file to read coadds from!")
        coadds = 0
    else:
        coadds = read_set_coadds(set_filename)
    
    if not aeri_pos_filename:
        LOG.error("no AERI.POS filename provided -> no position information")
        # FUTURE: we should short-cut AERI.POS with a direct Gps.py readout state?? 
        lat, lon, alt = nan, nan, nan
    else:
        lat, lon, alt = read_aeri_pos(aeri_pos_filename)
    
    return {'AERIunitNumber': model, 
            'instrumentUnitNumber': serial, 
            'systemReleaseNumber': nan,   # is in AERIHK.CFG which we retire for EAERI
            'coadditionsCount': coadds,  # number of co-adds specified in .SET file
            'sceneMirPosCount': nan,  # number of positions in scene mirror schedule
            'Latitude': lat,
            'Longitude': lon,
            'Altitude': alt
            }

def write_file(filename, scalars, vectors, max_vector_length, table = VECTOR_TABLE, history=''):
    """Write a file using the DMV OHWHIO module, containing the housekeeping scalars and vectors given.
    Set the file attributes correctly and informatively.
    
    """
    if os.path.exists(filename):
        raise EnvironmentError, 'file %s already exists' % filename
    fp = dmv.dmv()
    # reserve storage space for scalar housekeeping values
    fp.addHousekeepingSubheader()
    # set file metadata
    fp.setApplication('txt2hkspc')
    fp.setVersion('$Id: txt2hkspc.py,v 1.1.2.5 2008/05/15 20:30:31 rayg Exp $')
    fp.setHistory(history)
    # reserve storage space for high-resolution housekeeping
    name2vid = {}
    vid = 1
    for vec_sname, vec_ename, vec_label, vec_lname, handler, vec_units, vec_precision in table:
        # print vec_ename
        LOG.debug("creating %s (%s) (%s, %s)" % (vec_ename, vec_sname or '', vec_label, vec_units)) 
        fp.newVector(vid, 1.0, float(max_vector_length), max_vector_length)
        fp.setVectorShortname(vec_ename)
        fp.setVectorLongname(vec_lname or vec_sname or vec_ename)
        fp.setVectorLabel(vec_label or vec_sname or vec_ename)
        fp.setVectorUnits(vec_units or 'unknown')
        fp.setVectorPrecision(0,vec_precision or -3)
        fp.commitNewVector()
        name2vid[vec_ename] = vid
        vid += 1        
    rc = fp.createFile(filename)
    assert(rc==0)
    # get ids for scalars
    scalar_tmp = [(fp.findMetaID(key),val) for (key,val) in scalars.items()]
    scalar_ids = dmv.ints([x[0] for x in scalar_tmp])
    scalar_vals = dmv.floats([x[1] for x in scalar_tmp])
    # create a new record
    rc = fp.newRecord(1)
    assert(rc==0)
    # add the high resolution housekeeping
    for name,vod in name2vid.items():
        value = vectors[name]
        if len(value)<max_vector_length: 
            LOG.warning("%s has only %d values!" % (name,len(value)))
        assert(len(value)<=max_vector_length)
        # note that there may be fewer values in 
        rc = fp.addVector(name2vid[name], dmv.floats(value))
        assert(rc==0)
    # add the scalars
    rc = fp.addMeta(scalar_ids, scalar_vals)
    assert(rc==0)
    rc = fp.commitNewRecord()
    assert(rc==0)
    # file is implicitly closed on destructor
    
def test_write_file():
    # table of vectors to build
    vexation = [ ('HBBapexTemp', 'HBBapexTemp', 'HBB Apex Temp', None, None, 'K', -4), 
                ('HBBtopTemp', 'HBBtopTemp', 'HBB Top Temp', None, None, None, 0),
                ('HBBbottomTemp', 'HBBbottomTemp', 'HBB Bottom Temp', None, None, None, 0)
                ]
    scatest = { 'HBBapexTemp': -333., 'HBBtopTemp': 111., 'HBBbottomTemp': 222. }
    vectest = { 'HBBapexTemp': range(12), 
                'HBBtopTemp': range(12),
                'HBBbottomTemp': range(12)
                }
    TESTFN='/tmp/testhk.spc'
    if os.path.exists(TESTFN):
        os.unlink(TESTFN)
    write_file(TESTFN, scatest, vectest, 14, vexation)

def txt2hkspc(output_filename, input_filename, set_filename=None, inst_id_filename=None, aeri_pos_filename=None):
    """ write a *HK.SPC file from an EAERI .hk file
    Scalars are extracted from "high resolution" input data using a variety of functions.
    set_filename is the path of the current .SET file used by acqeaeri.
    """ 
    # read the input file into a dictionary of arrays
    txt = file(input_filename).read()
    hires = read_txt(txt)
    # extract the scalars
    scalars = extract_scalars(hires)
    LOG.debug('extracted %d scalars from hi-res data' % len(scalars))
    # read assorted configuration and output file values
    scalars.update( config_scalars(set_filename,inst_id_filename,aeri_pos_filename) )  
    LOG.debug('%s' % str(scalars))
    # determine the max vector length
    mvl = max( [len(q) for q in hires.values()] )
    LOG.debug("max vector length is %d" % mvl)
    LOG.debug('%s' % dict( [ (name,len(values)) for name,values in hires.items() if len(values)==mvl ] ))
    LOG.debug('%s' % dict( [ (name,len(values)) for name,values in hires.items() if len(values)!=mvl ] ))
    LOG.debug('writing the output file %s' % output_filename)    
    write_file(output_filename, scalars, hires, mvl, history = HISTORY_FMT % input_filename)
    LOG.debug('done')
    return hires

if __name__ == '__main__':
    logging.basicConfig()
    LOG.setLevel(logging.DEBUG)
    if len(sys.argv)==1: 
        LOG.info('running self-test') 
        test_write_file()
        print "usage: txt2hkspc outputhk.spc /path/to/hkdata.hk /path/to/VNNN.SET" 
    else:
        import Config
        D = txt2hkspc(sys.argv[1], sys.argv[2], set_filename=sys.argv[3], inst_id_filename = Config.INST_ID)
        # from pprint import pprint
        # pprint(', '.join(sorted(D.keys())))
        # pprint(D['HHMMSS'])
